#include "tm1637.h" 
#include "common.h"

#define FFFDEL 10
#define set_clklo() GPIOB->BSRR = GPIO_BSRR_BR_4; // reset
#define set_clkhi() GPIOB->BSRR = GPIO_BSRR_BS_4; // set
#define set_diolo() GPIOB->BSRR = GPIO_BSRR_BR_5; // reset
#define set_diohi() GPIOB->BSRR = GPIO_BSRR_BS_5; // set

void delayus(uint32_t i)
{
  for (; i > 0; i--) {
    for (uint32_t j = 0; j < FFFDEL; ++j) {
      __asm__ __volatile__("nop\n\t" ::
                               : "memory");
    }
  }
}

void set_dio_output (void)
{
  // двухтактный выход 10MHz 
  GPIOB->MODER   |=  GPIO_MODER_MODER5_0;
  GPIOB->OTYPER  &= ~GPIO_OTYPER_OT_5;
  GPIOB->OSPEEDR |=  GPIO_OSPEEDER_OSPEEDR5_0;
}

void set_clk_output (void)
{
  //двухтактный выход 10MHz 
  GPIOB->MODER   |=  GPIO_MODER_MODER4_0;
  GPIOB->OTYPER  &= ~GPIO_OTYPER_OT_4;
  GPIOB->OSPEEDR |=  GPIO_OSPEEDER_OSPEEDR4_0;
}

void set_dio_input_p (void)
{
  //вход с подтяжкой к +
  GPIOB->MODER &= ~GPIO_MODER_MODER5;
  GPIOB->PUPDR |=  GPIO_PUPDR_PUPDR5_0;
}

void set_dio_input_m(void)
{
  //вход с подтяжкой к -
  GPIOB->MODER &= ~GPIO_MODER_MODER5;
  GPIOB->PUPDR |=  GPIO_PUPDR_PUPDR5_1;
}

uint8_t set_dio_read(void)
{
 return (GPIOB->IDR & (uint32_t)(1<<5));
}

void tm_init(void)
{
  set_clk_output();
  set_dio_output();
  tm_set_brightness(2);
}

/*
#define TM1637_I2C_COMM1    0x40
#define TM1637_I2C_COMM2    0xC0
#define TM1637_I2C_COMM3    0x80
/
//      A
//     ---
//  F |   | B
//     -G-
//  E |   | C
//     ---
//      D
{
*/

#define emptyseg  0b00000000
#define minusseg  0b01000000
//static const uint8_t emptyseg  = 0b00000000;
//static const uint8_t minusseg  = 0b01000000;
//static const uint8_t degreeseg = 0b01100011;
const char seg_map[] =
{
    0x3f, 0x06, 0x5b, 0x4f, 0x66, 0x6d, 0x7d, 0x07, // 0-7
    0x7f, 0x6f, 0x77, 0x7c, 0x39, 0x5e, 0x79, 0x71, // 8-9, A-F
    0x00
};
// тоже самое
const uint8_t digitToSegment[] =
{
 // XGFEDCBA
  0b00111111,    // 0
  0b00000110,    // 1
  0b01011011,    // 2
  0b01001111,    // 3
  0b01100110,    // 4
  0b01101101,    // 5
  0b01111101,    // 6
  0b00000111,    // 7
  0b01111111,    // 8
  0b01101111,    // 9
  0b01110111,    // A
  0b01111100,    // b
  0b00111001,    // C
  0b01011110,    // d
  0b01111001,    // E
  0b01110001     // F
  };
// тоже 
const uint8_t digitHEX[] = 
{
  0x3f, 0x06, 0x5b, 0x4f,
  0x66, 0x6d, 0x7d, 0x07,
  0x7f, 0x6f, 0x00, 0x40
};//0~9, ,-

struct 
{
  uint8_t  tens, hundreds, thousands;
  uint16_t units;
} bcd4x;

struct 
{
  uint8_t units, tens;
} bcd2x;

static unsigned char digit_arr[5];

void tm_start(void) 
{
  set_clkhi();
  set_diohi();
  delayus(2);
  set_diolo();
}

void tm_stop(void) 
{
  set_clklo();
  delayus(2);
  set_diolo();
  delayus(2);
  set_clkhi();
  delayus(2);
  set_diohi();
}

void tm_read(void) 
{
  set_clklo();
  delayus(5);
  // while (dio); // We're cheating here and not actually reading back the response.
  set_clkhi();
  delayus(2);
  set_clklo();
}

void tm_wri(unsigned char b)
{
  for (int i = 0; i < 8; ++i) {
    set_clklo();
    if (b & 0x01) {
      set_diohi();
    } else {
      set_diolo();
    }
    delayus(3);
    b >>= 1;
    set_clkhi();
    delayus(3);
  }
}

// Valid brightness values: 0 - 8.
// 0 = display off.
void tm_set_brightness(char brightness)
{
  // Brightness command:
  // 1000 0XXX = display off
  // 1000 1BBB = display on, brightness 0-7
  // X = don't care
  // B = brightness
  tm_start();
  tm_wri(0x87 + brightness);
  tm_read();
  tm_stop();
}

void tm_send4x()
{
  tm_start();
  tm_wri(0x40);
  tm_read();
  tm_stop();
  tm_start();
  tm_wri(0xc0);
  tm_read();
  for (int i = 0; i < 4; ++i) {
    tm_wri(digit_arr[3 - i]);
    tm_read();
  }
  tm_stop();
}

void tm_disp_clear(void)
{
  for (int i = 0; i < 4; ++i) {
    digit_arr[i] = emptyseg;
  }
  tm_send4x();
}

// Раскладываем 9999 на цифры
void bin_bcd4x(uint_fast16_t a)
{
  bcd4x.tens = 0;
  bcd4x.hundreds = 0;
  bcd4x.thousands = 0;
  bcd4x.units = a;
  while(bcd4x.units >= 1000) {
    bcd4x.units -= 1000;
    bcd4x.thousands++;
  }
  while(bcd4x.units >= 100) {
    bcd4x.units -= 100;
    bcd4x.hundreds++;
  }
  while(bcd4x.units >= 10) {
    bcd4x.units -= 10;
    bcd4x.tens++;
  }
}

void bin_bcd2x(uint_fast8_t a)
{
  bcd2x.tens = 0;
  bcd2x.units = a;
  while(bcd2x.units >= 10) {
    bcd2x.units -= 10;
    bcd2x.tens++;
  }
}

void tm_disp_dec(uint16_t x, uint8_t colon)
{
  for (int i = 0; i < 4; ++i) {
    digit_arr[i] = seg_map[x % 10];
    if (i == 2 && colon) {
      digit_arr[i] |= 1 << 7;
    }
      x /= 10;
  }
  tm_send4x();
}

// XXXX с лидирующими нолями, с двоеточием
void tm_disp_decx(uint16_t x, uint8_t point)
{
  bin_bcd4x(x);
  
  digit_arr[3] = seg_map[bcd4x.thousands];

  if (point)
  { digit_arr[2] = seg_map[bcd4x.hundreds]; digit_arr[2] |= 1 << 7; }
  else digit_arr[2] = seg_map[bcd4x.hundreds]; 

  digit_arr[1] = seg_map[bcd4x.tens];
  digit_arr[0] = seg_map[bcd4x.units];
  tm_send4x();
}

// 9999 без лидирующих нолей
void tm_disp_decz(uint16_t x)
{
  bin_bcd4x(x);
  
  if  (bcd4x.thousands == 0)      digit_arr[3] = emptyseg; 
  else digit_arr[3] = seg_map[bcd4x.thousands];
  
  if ((bcd4x.hundreds  == 0) && 
      (bcd4x.thousands == 0))     digit_arr[2] = emptyseg; 
  else digit_arr[2] = seg_map[bcd4x.hundreds];
  
  if ((bcd4x.tens      == 0) && 
      (bcd4x.thousands == 0) &&
      (bcd4x.hundreds  == 0))     digit_arr[1] = emptyseg; 
  else digit_arr[1] = seg_map[bcd4x.tens];
  
  digit_arr[0] = seg_map[bcd4x.units];
  
  tm_send4x();
}

// XXXX без лидирующих нолей, с двоеточием
void tm_disp_decA(uint16_t x, uint8_t point)
{
  bin_bcd4x(x);
  if (bcd4x.thousands == 0) digit_arr[3] = emptyseg; else digit_arr[3] = seg_map[bcd4x.thousands];
  if ((bcd4x.hundreds  == 0) && (bcd4x.thousands == 0)) digit_arr[2] = emptyseg;
  else {
    if (point) { 
      digit_arr[2] = seg_map[bcd4x.hundreds];
      digit_arr[2] |= 1 << 7;
    } else digit_arr[2] = seg_map[bcd4x.hundreds]; 
  }
  if ((bcd4x.tens == 0) && (bcd4x.thousands == 0) && (bcd4x.hundreds == 0)) digit_arr[2] = emptyseg; else digit_arr[1] = seg_map[bcd4x.tens];
  digit_arr[0] = seg_map[bcd4x.units];
  tm_send4x();
}

// displays [XX]
void tm_disp_channel(uint8_t x)
{
  bin_bcd4x(x);
  digit_arr[3] = 0b00111001;
  if (bcd4x.tens == 0) {
  digit_arr[2] = emptyseg;
  } else
  digit_arr[2] = seg_map[bcd4x.tens];
  digit_arr[1] = seg_map[bcd4x.units];
  digit_arr[0] = 0b00001111;
  tm_send4x();
}

void tm_disp_chsaved(uint8_t x)
{
  bin_bcd4x(x);
  digit_arr[3] = minusseg;
  if (bcd4x.tens == 0) {
  digit_arr[2] = emptyseg;
  } else
  digit_arr[2] = seg_map[bcd4x.tens];
  digit_arr[1] = seg_map[bcd4x.units];
  digit_arr[0] = minusseg;
  tm_send4x();
}

// displays voXX
void tm_disp_volume(uint8_t x, uint8_t bass) 
{
  bin_bcd4x(x);
  digit_arr[3] = 0b00011100;
  if (!bass) digit_arr[2] = 0b01011100; else digit_arr[2] = 0b01111100;
  if (bcd4x.tens == 0) {
    digit_arr[1] = emptyseg;
  } else
  digit_arr[1] = seg_map[bcd4x.tens];
  digit_arr[0] = seg_map[bcd4x.units];
  tm_send4x();
}

// displays _oo_
void tm_disp_oo(void)
{
  digit_arr[3] = emptyseg; // clear
  digit_arr[2] = 0b01011100; // o 
  digit_arr[1] = 0b01011100; // o
  digit_arr[0] = emptyseg; // clear
  tm_send4x();
}

void tm_disp_oops(void)
{
  digit_arr[3] = emptyseg; // clear
  digit_arr[2] = 0b00001111; // o 
  digit_arr[1] = 0b00111001; // o
  digit_arr[0] = emptyseg; // clear
  tm_send4x();
}

// тоже, что и, но void tm_disp_decz(uint16_t x) {}
void tm_disp_bcd10freq(uint16_t x)
{
  bin_bcd4x(x);
  
  if  (bcd4x.thousands == 0)      digit_arr[3] = emptyseg; 
  else digit_arr[3] = seg_map[bcd4x.thousands];
  
  if ((bcd4x.hundreds  == 0) && 
      (bcd4x.thousands == 0))     digit_arr[2] = emptyseg; 
  else digit_arr[2] = seg_map[bcd4x.hundreds]; // | 0x80;
  
  if ((bcd4x.tens      == 0) && 
      (bcd4x.thousands == 0) &&
      (bcd4x.hundreds  == 0))     digit_arr[1] = emptyseg; 
  else digit_arr[1] = seg_map[bcd4x.tens];
  
  digit_arr[0] = seg_map[bcd4x.units];
  
  tm_send4x();
}

// Набрёл в тему https://we.easyelectronics.ru/blog/Soft/2400.html
char* shift_and_mul_utoa16(uint16_t n, char *buffer)
{
    uint8_t d4, d3, d2, d1, q, d0;

        d1 = (n>>4)  & 0xF;
        d2 = (n>>8)  & 0xF;
        d3 = (n>>12) & 0xF;

        d0 = 6*(d3 + d2 + d1) + (n & 0xF);
        q = (d0 * 0xCD) >> 11;
        d0 = d0 - 10*q;

        d1 = q + 9*d3 + 5*d2 + d1;
        q = (d1 * 0xCD) >> 11;
        d1 = d1 - 10*q;

        d2 = q + 2*d2;
        q = (d2 * 0x1A) >> 8;
        d2 = d2 - 10*q;

        d3 = q + 4*d3;
        d4 = (d3 * 0x1A) >> 8;
        d3 = d3 - 10*d4;

        char *ptr = buffer;
    *ptr++ = ( d4 + '0' );
    *ptr++ = ( d3 + '0' );
    *ptr++ = ( d2 + '0' );
    *ptr++ = ( d1 + '0' );
    *ptr++ = ( d0 + '0' );
        *ptr = 0;

        while(buffer[0] == '0') ++buffer;
        return buffer;
}

//
void tm_disp_div10freq(uint16_t x)
{
  char _arr[5];
  
  shift_and_mul_utoa16( x, _arr);
  
  if (_arr[0] == '0') digit_arr[3] = emptyseg;
  else digit_arr[3] = seg_map[(uint8_t)(_arr[0] - '0')];
  
  digit_arr[2] = seg_map[(uint8_t)(_arr[1] - '0')];
  digit_arr[1] = seg_map[(uint8_t)(_arr[2] - '0')];
  digit_arr[0] = seg_map[(uint8_t)(_arr[3] - '0')];
  
  tm_send4x();
}

void tm_disp_clock(uint8_t hour, uint8_t min, uint8_t colon)
{
  bin_bcd2x(hour);
  if (!bcd2x.tens) digit_arr[3] = emptyseg; else digit_arr[3] = seg_map[bcd2x.tens];
  if (colon) digit_arr[2] = seg_map[bcd2x.units] | 0x80; else digit_arr[2] = seg_map[bcd2x.units];
  
  bin_bcd2x(min);
  digit_arr[1] = seg_map[bcd2x.tens]; //if (!bcd2x.tens) digit_arr[1] = emptyseg; else 
  digit_arr[0] = seg_map[bcd2x.units];


  tm_send4x();
}

//
//void shift_and_mul_16x(uint16_t n, uint8_t *_arr){}
//void tm_disp_10(uint16_t x) {uint8_t _arr[4];tm_send4x();}
//void tm_disp_10x(uint16_t x) {uint8_t _arr[4]; tm_send4x();}

/*
// ASCII Map - Index 0 starts at ASCII 32
const uint8_t asciiToSegment[] = {
   0b00000000, // 032 (Space) 0
   0b00110000, // 033 !       1
   0b00100010, // 034 "       2
   0b01000001, // 035 #       3
   0b01101101, // 036 $       4
   0b01010010, // 037 %       5
   0b01111100, // 038 &       6
   0b00100000, // 039 '
   0b00111001, // 040 (
   0b00001111, // 041 )
   0b00100001, // 042 *
   0b01110000, // 043 +
   0b00001000, // 044 ,
   0b01000000, // 045 -
   0b00001000, // 046 .
   0b01010010, // 047 /
   0b00111111, // 048 0
   0b00000110, // 049 1
   0b01011011, // 050 2
   0b01001111, // 051 3
   0b01100110, // 052 4
   0b01101101, // 053 5
   0b01111101, // 054 6
   0b00000111, // 055 7
   0b01111111, // 056 8
   0b01101111, // 057 9
   0b01001000, // 058 :
   0b01001000, // 059 ;
   0b00111001, // 060 <
   0b01001000, // 061 =
   0b00001111, // 062 >
   0b01010011, // 063 ?
   0b01011111, // 064 @
   0b01110111, // 065 A
   0b01111100, // 066 B
   0b00111001, // 067 C
   0b01011110, // 068 D
   0b01111001, // 069 E
   0b01110001, // 070 F
   0b00111101, // 071 G
   0b01110110, // 072 H
   0b00000110, // 073 I
   0b00011110, // 074 J
   0b01110110, // 075 K
   0b00111000, // 076 L
   0b00010101, // 077 M
   0b00110111, // 078 N
   0b00111111, // 079 O
   0b01110011, // 080 P
   0b01100111, // 081 Q
   0b00110001, // 082 R
   0b01101101, // 083 S
   0b01111000, // 084 T
   0b00111110, // 085 U
   0b00011100, // 086 V
   0b00101010, // 087 W
   0b01110110, // 088 X
   0b01101110, // 089 Y
   0b01011011, // 090 Z
   0b00111001, // 091 [
   0b01100100, // 092 (backslash)
   0b00001111, // 093 ]
   0b00100011, // 094 ^
   0b00001000, // 095 _
   0b00100000, // 096 `
   0b01110111, // 097 a
   0b01111100, // 098 b
   0b01011000, // 099 c
   0b01011110, // 100 d
   0b01111001, // 101 e
   0b01110001, // 102 f
   0b01101111, // 103 g
   0b01110100, // 104 h
   0b00000100, // 105 i
   0b00011110, // 106 j
   0b01110110, // 107 k
   0b00011000, // 108 l
   0b00010101, // 109 m
   0b01010100, // 110 n
   0b01011100, // 111 o
   0b01110011, // 112 p
   0b01100111, // 113 q
   0b01010000, // 114 r
   0b01101101, // 115 s
   0b01111000, // 116 t
   0b00011100, // 117 u
   0b00011100, // 118 v
   0b00101010, // 119 w
   0b01110110, // 120 x
   0b01101110, // 121 y
   0b01011011, // 122 z
   0b00111001, // 123 {
   0b00110000, // 124 |
   0b00001111, // 125 }
   0b01000000, // 126 ~
   0b00000000  // 127 
};
*/

//const unsigned char cDigit2Seg[] = {0x3f, 0x6, 0x5b, 0x4f, 0x66, 0x6d, 0x7d, 0x07, 0x7f, 0x6f};
/*
const uint8_t digitToSegment[] = {
 // XGFEDCBA
  0b00111111,    // 0
  0b00000110,    // 1
  0b01011011,    // 2
  0b01001111,    // 3
  0b01100110,    // 4
  0b01101101,    // 5
  0b01111101,    // 6
  0b00000111,    // 7
  0b01111111,    // 8
  0b01101111,    // 9
  0b01110111,    // A
  0b01111100,    // b
  0b00111001,    // C
  0b01011110,    // d
  0b01111001,    // E
  0b01110001     // F
  };
*/
/*
void tm_disp_dec(uint16_t x, uint8_t colon) {
  for (int i = 0; i < 4; ++i) {
    digit_arr[i] = seg_map[x % 10];
    if (i == 2 && colon) {
      digit_arr[i] |= 1 << 7;
    }
      x /= 10;
  }
  tm_start();
  tm_wri(0x40);
  tm_read();
  tm_stop();
  tm_start();
  tm_wri(0xc0);
  tm_read();
  for (int i = 0; i < 4; ++i) {
    tm_wri(digit_arr[3 - i]);
    tm_read();
  }
  tm_stop();
}
*/

// 3. Извлечение цифр делением на 10.
// Готовые стандартные и не очень способы посмотрели. Теперь пришло время свои велосипеды изобретать. 
// Первый самый очевидный способ это конечно деление на 10 и вычисление остатка в цикле.
char * utoa_builtin_div(uint32_t value, char *buffer)
{
   buffer += 11; 
// 11 байт достаточно для десятичного представления 32-х байтного числа
// и  завершающего нуля
   *--buffer = 0;
   do
   {
      *--buffer = value % 10 + '0';
      value /= 10;
   }
   while (value != 0);
   return buffer;
}

// Аутсайдером везде является функция использующая библиотечную функцию деления div. Несмотря на то, что она возвращает за один вызов и остаток и частное 
// от деления, даже на STM32 аппаратным делением, она реализована программно и работает очень медленно. Очевидно этот способ использовать не стоит. 
// Однако функция использующая встроенный оператор деления utoa_builtin_div, плетущаяся в конце на AVR и MSP430, на STM32 — в лидерах. 
// Ничего удивительного, ведь в Cortex M3 есть аппаратное деление скажут многие, и будут не совсем правы — деление-то там есть, но оно не такое уж и быстрое 
// (в скобках для utoa_builtin_div указано время, если заставить компилятор сгенерировать честное деление). Дело в том, что хитрый GCC при делении на константу 
// использует хитрый трюк — заменяет деление на умножение на константу такую, что старшие 32 бита в 64 разрядном произведении, содержат исходное делимое, 
// делённое на 10. 

const uint32_t  pow10Table32[]=
{
    1000000000ul,
    100000000ul,
    10000000ul,
    1000000ul,
    100000ul,
    10000ul,
    1000ul,
    100ul,
    10ul,
    1ul
};

static char *utoa_cycle_sub(uint32_t value, char *buffer)
{
   if(value == 0)
   {
      buffer[0] = '0';
      buffer[1] = 0;
      return buffer;
   }
   char *ptr = buffer;
   uint8_t i = 0;
   do
   {
      uint32_t pow10 = pow10Table32[i++];
      uint8_t count = 0;
      while(value >= pow10)
      {
         count ++;
         value -= pow10;
      }
      *ptr++ = count + '0';
   }while(i < 10);
   *ptr = 0;
// удаляем ведущие нули
   while(buffer[0] == '0') ++buffer;
   return buffer;
}